import uuid
from textwrap import dedent

from django.db import transaction
from django.shortcuts import get_object_or_404, redirect
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import serializers, status
from rest_framework.authentication import SessionAuthentication, TokenAuthentication
from rest_framework.exceptions import NotFound, ValidationError
from rest_framework.generics import CreateAPIView, ListCreateAPIView, RetrieveUpdateAPIView, UpdateAPIView
from rest_framework.response import Response
from rest_framework.views import APIView

from arkindex.ponos.models import FINAL_STATES, Artifact, State, Task, task_token_default
from arkindex.ponos.permissions import (
    IsAgentOrArtifactGuest,
    IsAgentOrTaskGuest,
    IsAssignedAgentOrReadOnly,
    IsAssignedAgentOrTaskOrReadOnly,
    IsTaskAdmin,
)
from arkindex.ponos.serializers import ArtifactSerializer, TaskSerializer, TaskTinySerializer
from arkindex.project.mixins import ProcessACLMixin
from arkindex.project.permissions import IsVerified
from arkindex.users.models import Role


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveTaskFromAgent",
        description=dedent("""
        Retrieve a Ponos task.

        Requires **guest** access on the task's process, Ponos agent authentication, or authentication as the Ponos task to retrieve.
        """),
    ),
    put=extend_schema(
        operation_id="UpdateTaskFromAgent",
        description=dedent("""
        Update a task.

        Requires authentication as the Ponos agent assigned to the task.
        """),
    ),
    patch=extend_schema(
        operation_id="PartialUpdateTaskFromAgent",
        description=dedent("""
        Partially update a task.

        Requires authentication as the Ponos agent assigned to the task.
        """),
    ),
)
class TaskDetailsFromAgent(RetrieveUpdateAPIView):

    # Avoid stale read when a recently assigned agent wants to update
    # the state of one of its tasks
    queryset = Task.objects.using("default").select_related(
        "agent__farm",
        "gpu",
        # Used for permission checks
        "process__corpus"
    )
    permission_classes = (
        # On all HTTP methods, require either any Ponos agent, an instance admin, the task itself, or guest access to the process' task
        IsAgentOrTaskGuest,
        # On unsafe HTTP methods, require a Ponos agent assigned to the task. Both permission classes are combined.
        IsAssignedAgentOrReadOnly,
    )
    serializer_class = TaskSerializer


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListArtifacts",
        description=dedent("""
        List the artifacts of a task.

        Requires **guest** access on the task's process, Ponos agent authentication, or authentication as the Ponos task to list artifacts of.
        """),
    ),
    post=extend_schema(
        operation_id="CreateArtifact",
        description=dedent("""
        Create an artifact on task.

        Requires authentication as the Ponos task to create an artifact on, or as a Ponos agent assigned to the task.
        """),
    ),
)
class TaskArtifacts(ListCreateAPIView):

    # Used for OpenAPI schema serialization: the ID in the path is the task ID
    queryset = Task.objects.none()
    serializer_class = ArtifactSerializer
    permission_classes = (
        # On all HTTP methods, require either any Ponos agent, an instance admin, the task itself, or guest access to the process' task
        IsAgentOrTaskGuest,
        # On unsafe HTTP methods, require a Ponos agent assigned to the task, or the task itself. Both permission classes are combined.
        IsAssignedAgentOrTaskOrReadOnly,
    )

    # Force no pagination, even when global settings add them
    pagination_class = None

    @property
    def task(self):
        task = get_object_or_404(
            # Select the required tables for permissions checking
            Task.objects.select_related("process__corpus"),
            pk=self.kwargs["pk"],
        )
        self.check_object_permissions(self.request, task)
        return task

    def get_queryset(self):
        return self.task.artifacts.all()

    def perform_create(self, serializer):
        # Assign task when creating through the API
        serializer.save(task=self.task)


class TaskArtifactDownload(APIView):
    """
    Redirects to the S3 URL of an artifact in order to download it.
    """

    permission_classes = (IsAgentOrArtifactGuest, )

    def get_object(self, pk, path):
        artifact = get_object_or_404(
            # Select the required tables for permissions checking
            Artifact.objects.select_related("task__process__corpus"),
            task_id=pk,
            path=path,
        )
        self.check_object_permissions(self.request, artifact)
        return artifact

    @extend_schema(
        operation_id="DownloadArtifact",
        responses={302: None},
        tags=["ponos"],
    )
    def get(self, request, *args, **kwargs):
        artifact = self.get_object(*args, **kwargs)
        return redirect(artifact.s3_url)


@extend_schema(tags=["ponos"])
@extend_schema_view(
    put=extend_schema(
        description=dedent("""
        Update a task.

        Requires **admin** access on the task's process, or to be the creator of the process.
        Cannot be used with Ponos agent or task authentication.
        """),
    ),
    patch=extend_schema(
        description=dedent("""
        Partially update a task.

        Requires **admin** access on the task's process, or to be the creator of the process.
        Cannot be used with Ponos agent or task authentication.
        """),
    ),
)
class TaskUpdate(UpdateAPIView):
    # Only allow regular users, not Ponos agents or tasks
    authentication_classes = (TokenAuthentication, SessionAuthentication)
    # Only allow regular users that have admin access to the task's process
    permission_classes = (IsTaskAdmin, )
    queryset = Task.objects.select_related("process__corpus")
    serializer_class = TaskTinySerializer


@extend_schema_view(
    post=extend_schema(
        operation_id="RestartTask",
        tags=["ponos"],
        description=dedent(
            """
            Restart a task by creating a fresh copy and moving dependent tasks to the new one.

            Scenario restarting `my_worker` task:
            ```
            init_elements → my_worker → other worker
            ```
            ```
            init_elements → my_worker
                       ↘
                         my_worker_2 → other worker
            ```

            Requires an **admin** access to the task's process.
            The task must be in a final state to be restarted.
            """
        ),
        responses={201: TaskSerializer},
    ),
)
class TaskRestart(ProcessACLMixin, CreateAPIView):
    permission_classes = (IsVerified,)
    serializer_class = serializers.Serializer

    def get_task(self):
        task = get_object_or_404(
            Task.objects.prefetch_related("parents").select_related("process__corpus"),
            pk=self.kwargs["pk"],
        )
        access_level = self.process_access_level(task.process)
        if access_level is None:
            raise NotFound
        if access_level < Role.Admin.value:
            raise ValidationError(
                detail="You do not have an admin access to the process of this task."
            )
        if task.state not in FINAL_STATES:
            raise ValidationError(
                detail="Task's state must be in a final state to be restarted."
            )
        # TODO Check the original_task_id field directly once it is implemented
        # https://gitlab.teklia.com/arkindex/frontend/-/issues/1383
        if task.process.tasks.filter(run=task.run, slug=self.increment(task.slug)).exists():
            raise ValidationError(
                detail="This task has already been restarted"
            )
        return task

    def increment(self, name):
        basename, *suffix = name.rsplit("_restart", 1)
        suffix = int(suffix[0]) + 1 if suffix and suffix[0].isdigit() else 1
        return f"{basename}_restart{suffix}"

    @transaction.atomic
    def create(self, request, pk=None, **kwargs):
        copy = self.get_task()
        parents = list(copy.parents.all())

        copy.id = uuid.uuid4()
        copy.state = State.Pending
        copy.token = task_token_default()
        copy.slug = self.increment(copy.slug)
        copy.save()

        # Create links to retried task parents
        copy.parents.add(*parents)

        # Move all tasks depending on the retried task to the copy
        Task.children.through.objects.filter(to_task_id=pk).update(to_task_id=copy.id)

        return Response(TaskSerializer(copy).data, status=status.HTTP_201_CREATED)
